﻿using LAViD.Structures;
using LAViD.Unity.Utils;
using LAViD.Utils;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using UnityEngine;

namespace LAViD.VLibras.Dictionary
{
	public class Dictionary : MonoBehaviour
	{
		private readonly string DefaultDictionary = "{}"; // "{\"keys\": {}, \"trie\": {}, \"characters\": []}";
		private readonly char[] CharactersSigns = new char[] {
			'A', 'B', 'C', 'Ç', 'D', 'E', 'F', 'G', 'H', 'I',
			'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S',
			'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
			'0', '1', '2', '3', '4', '5', '6', '7', '8', '9'
		};

		private string path;
		private string versionPath;

		private string version = "0.0.0";
		private Trie signs = null;
		private volatile bool ready = false;
		private event Events.OnLoad onLoadListeners;

		public void AddOnLoadListener(Events.OnLoad onLoad)
		{
			PlayerLogger.Log("D", "AOLL", "Registering onLoad event.");

			if (ready)
			{
				PlayerLogger.Log("D", "AOLL", "Ready. Calling onLoad event directly.");
				onLoad();
			}
			else this.onLoadListeners += onLoad;
		}

		public Trie Signs
		{
			get { return ready ? this.signs : null; }
		}

		private void Start()
		{
			this.path = Application.persistentDataPath + "/dictionary.json";
			this.versionPath = Application.persistentDataPath + "/dictionary_version.txt";

			PlayerLogger.Log("D", "S", "Dictionary path: " + this.path);
			PlayerLogger.Log("D", "S", "Dictionary version path: " + this.versionPath);

			if (!File.Exists(this.versionPath))
			{
				File.WriteAllText(this.versionPath, version);
				save(DefaultDictionary);
			}
			else this.version = File.ReadAllText(this.versionPath);

			StartCoroutine(Methods.WaitForResponse(
				new WWW("http://atualizacao.vlibras.lavid.ufpb.br/version/dict"),
				10f,
				versionRequestSuccess,
				delegate (WWW response)
				{
					PlayerLogger.Log("D", "S", "Connection error. Loading local dictionary version " + this.version + ".");

					if (response.error == null)
						PlayerLogger.Log("D", "S", "Timeout.");
					else
						PlayerLogger.Log("D", "S", "WWW error: " + response.error);

					load(read());
				}
			));
		}

		private void Update()
		{
			if (ready)
			{
				PlayerLogger.Log("D", "AOLL", "Ready. Calling onLoad events.");

				this.onLoadListeners();
                this.gameObject.GetComponent<Dictionary>().enabled = false;
			}
		}

		private void versionRequestSuccess(WWW response)
		{
			PlayerLogger.Log("D", "vRS", "Version request done.");
			string newVersion = response.text;

			if (isUpdated(newVersion))
			{
				PlayerLogger.Log("D", "vRS", "Version (" + this.version + ") is up to date. Loading local dictionary.");
				load(read());
			}
			else
			{
				PlayerLogger.Log("D", "vRS", "Updating dictionary from version " + this.version + " to " + newVersion + ".");
				File.WriteAllText(this.versionPath, newVersion);

				StartCoroutine(Methods.WaitForResponse(
					new WWW("http://dicionario.vlibras.gov.br/signs"),
					20f,
					delegate (WWW jsonResponse)
					{
						PlayerLogger.Log("D", "vRS", "Trie request done.");
						string json = jsonResponse.text;

						save(json);
						load(json);
					},
					delegate (WWW jsonResponse)
					{
						if (jsonResponse.error == null)
							PlayerLogger.Log("D", "vRS", "Timeout.");
						else
							PlayerLogger.Log("D", "vRS", "WWW error: " + jsonResponse.error);

						load(read());
					}
				));
			}
		}

		private bool isUpdated(string newVersion)
		{
			string[] versionTokens = this.version.Split('.');
			string[] newVersionTokens = newVersion.Split('.');
			
			for (int i = 0; i < versionTokens.Length; i++)
				if (int.Parse(versionTokens[i]) < int.Parse(newVersionTokens[i]))
					return false;

			return true;
		}

		private void load(string json)
		{
			new Thread(
				() => {
					Thread.CurrentThread.IsBackground = true;

					Stopwatch watch = new Stopwatch();
					watch.Start();

					try {
						PlayerLogger.Log("D", "l", "Starting trie parsing.");
						this.signs = Trie.FromJSON(new JSONParser(Regex.Unescape(json))); ;
							
						foreach (char c in CharactersSigns)
							this.signs.Add(c.ToString());
					}
					catch (Exception e) {
						PlayerLogger.Log("D", "l", "Exception at trie JSON parser: " + e.ToString());
						this.signs = new Trie();
					}

					this.ready = true;

					watch.Stop();
					PlayerLogger.Log("D", "l", "Parsed in " + watch.ElapsedMilliseconds + " ms.");
				}
			).Start();
		}

		private bool exists() {
			return File.Exists(this.path);
		}

		private string read()
		{
			if (!exists())
			{
				PlayerLogger.Log("D", "r", "Error! Local dictionary not found. Loading default.");
				return DefaultDictionary;
			}

			return File.ReadAllText(this.path);
		}

		private void save(string json)
		{
			File.WriteAllText(this.path, json);

#if UNITY_EDITOR
			UnityEditor.AssetDatabase.Refresh();
#endif
		}
	}
}
